# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - EventAggregator library

    @copyright: 2008, 2009, 2010, 2011, 2012 by Paul Boddie <paul@boddie.org.uk>
    @copyright: 2000-2004 Juergen Hermann <jh@web.de>,
                2005-2008 MoinMoin:ThomasWaldmann.
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from LocationSupport import *
from MoinDateSupport import *
from MoinSupport import *

from MoinMoin.Page import Page
from MoinMoin.action import cache
from MoinMoin import caching
from MoinMoin import search
from MoinMoin import wikiutil

import codecs
import time
import re
import urllib, urllib2

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

try:
    set
except NameError:
    from sets import Set as set

try:
    import vCalendar
except ImportError:
    vCalendar = None

escape = wikiutil.escape

__version__ = "0.9"

# Regular expressions where MoinMoin does not provide the required support.

category_regexp = None

# Page parsing.

definition_list_regexp = re.compile(ur'(?P<wholeterm>^(?P<optcomment>#*)\s+(?P<term>.*?):: )(?P<desc>.*?)$', re.UNICODE | re.MULTILINE)
category_membership_regexp = re.compile(ur"^\s*(?:(Category\S+)(?:\s+(Category\S+))*)\s*$", re.MULTILINE | re.UNICODE)

# Value parsing.

country_code_regexp = re.compile(ur'(?:^|\W)(?P<code>[A-Z]{2})(?:$|\W+$)', re.UNICODE)

# Simple content parsing.

verbatim_regexp = re.compile(ur'(?:'
    ur'<<Verbatim\((?P<verbatim>.*?)\)>>'
    ur'|'
    ur'\[\[Verbatim\((?P<verbatim2>.*?)\)\]\]'
    ur'|'
    ur'!(?P<verbatim3>.*?)(\s|$)?'
    ur'|'
    ur'`(?P<monospace>.*?)`'
    ur'|'
    ur'{{{(?P<preformatted>.*?)}}}'
    ur')', re.UNICODE)

# Utility functions.

def getCategoryPattern(request):
    global category_regexp

    try:
        return request.cfg.cache.page_category_regexact
    except AttributeError:

        # Use regular expression from MoinMoin 1.7.1 otherwise.

        if category_regexp is None:
            category_regexp = re.compile(u'^%s$' % ur'(?P<all>Category(?P<key>(?!Template)\S+))', re.UNICODE)
        return category_regexp

def getWikiDict(pagename, request):
    if Page(request, pagename).exists() and request.user.may.read(pagename):
        if hasattr(request.dicts, "dict"):
            return request.dicts.dict(pagename)
        else:
            return request.dicts[pagename]
    else:
        return None

def to_list(s, sep):
    return [x.strip() for x in s.split(sep) if x.strip()]

def sort_none_first(x, y):
    if x is None:
        return -1
    elif y is None:
        return 1
    else:
        return cmp(x, y)

def sort_start_first(x, y):
    x_ts = x.as_limits()
    if x_ts is not None:
        x_start, x_end = x_ts
        y_ts = y.as_limits()
        if y_ts is not None:
            y_start, y_end = y_ts
            start_order = cmp(x_start, y_start)
            if start_order == 0:
                return cmp(x_end, y_end)
            else:
                return start_order
    return 0

# Utility classes and associated functions.

class ActionSupport(ActionSupport):

    "Extend the generic action support."

    def get_month_lists(self, default_as_current=0):

        """
        Return two lists of HTML element definitions corresponding to the start
        and end month selection controls, with months selected according to any
        values that have been specified via request parameters.
        """

        _ = self._
        form = self.get_form()

        # Initialise month lists.

        start_month_list = []
        end_month_list = []

        start_month = self._get_input(form, "start-month", default_as_current and getCurrentMonth().month() or None)
        end_month = self._get_input(form, "end-month", start_month)

        # Prepare month lists, selecting specified months.

        if not default_as_current:
            start_month_list.append('<option value=""></option>')
            end_month_list.append('<option value=""></option>')

        for month in range(1, 13):
            month_label = escape(_(getMonthLabel(month)))
            selected = self._get_selected(month, start_month)
            start_month_list.append('<option value="%02d" %s>%s</option>' % (month, selected, month_label))
            selected = self._get_selected(month, end_month)
            end_month_list.append('<option value="%02d" %s>%s</option>' % (month, selected, month_label))

        return start_month_list, end_month_list

    def get_year_defaults(self, default_as_current=0):

        "Return defaults for the start and end years."

        form = self.get_form()

        start_year_default = form.get("start-year", [default_as_current and getCurrentYear() or ""])[0]
        end_year_default = form.get("end-year", [default_as_current and start_year_default or ""])[0]

        return start_year_default, end_year_default

    def get_day_defaults(self, default_as_current=0):

        "Return defaults for the start and end days."

        form = self.get_form()

        start_day_default = form.get("start-day", [default_as_current and getCurrentDate().day() or ""])[0]
        end_day_default = form.get("end-day", [default_as_current and start_day_default or ""])[0]

        return start_day_default, end_day_default

# Textual representations.

def getSimpleWikiText(text):

    """
    Return the plain text representation of the given 'text' which may employ
    certain Wiki syntax features, such as those providing verbatim or monospaced
    text.
    """

    # NOTE: Re-implementing support for verbatim text and linking avoidance.

    return "".join([s for s in verbatim_regexp.split(text) if s is not None])

def getEncodedWikiText(text):

    "Encode the given 'text' in a verbatim representation."

    return "<<Verbatim(%s)>>" % text

def getPrettyTitle(title):

    "Return a nicely formatted version of the given 'title'."

    return title.replace("_", " ").replace("/", u"  ")

# Category discovery and searching.

def getCategories(request):

    """
    From the AdvancedSearch macro, return a list of category page names using
    the given 'request'.
    """

    # This will return all pages with "Category" in the title.

    cat_filter = getCategoryPattern(request).search
    return request.rootpage.getPageList(filter=cat_filter)

def getCategoryMapping(category_pagenames, request):

    """
    For the given 'category_pagenames' return a list of tuples of the form
    (category name, category page name) using the given 'request'.
    """

    cat_pattern = getCategoryPattern(request)
    mapping = []
    for pagename in category_pagenames:
        name = cat_pattern.match(pagename).group("key")
        if name != "Category":
            mapping.append((name, pagename))
    mapping.sort()
    return mapping

def getCategoryPages(pagename, request):

    """
    Return the pages associated with the given category 'pagename' using the
    'request'.
    """

    query = search.QueryParser().parse_query('category:%s' % pagename)
    results = search.searchPages(request, query, "page_name")

    cat_pattern = getCategoryPattern(request)
    pages = []
    for page in results.hits:
        if not cat_pattern.match(page.page_name):
            pages.append(page)
    return pages

def getAllCategoryPages(category_names, request):

    """
    Return all pages belonging to the categories having the given
    'category_names', using the given 'request'.
    """

    pages = []
    pagenames = set()

    for category_name in category_names:

        # Get the pages and page names in the category.

        pages_in_category = getCategoryPages(category_name, request)

        # Visit each page in the category.

        for page_in_category in pages_in_category:
            pagename = page_in_category.page_name

            # Only process each page once.

            if pagename in pagenames:
                continue
            else:
                pagenames.add(pagename)

            pages.append(page_in_category)

    return pages

def getPagesFromResults(result_pages, request):

    "Return genuine pages for the given 'result_pages' using the 'request'."

    return [Page(request, page.page_name) for page in result_pages]

# Event resources providing collections of events.

class EventResource:

    "A resource providing event information."

    def __init__(self, url):
        self.url = url

    def getPageURL(self):

        "Return the URL of this page."

        return self.url

    def getFormat(self):

        "Get the format used by this resource."

        return "plain"

    def getMetadata(self):

        """
        Return a dictionary containing items describing the page's "created"
        time, "last-modified" time, "sequence" (or revision number) and the
        "last-comment" made about the last edit.
        """

        return {}

    def getEvents(self):

        "Return a list of events from this resource."

        return []

    def linkToPage(self, request, text, query_string=None):

        """
        Using 'request', return a link to this page with the given link 'text'
        and optional 'query_string'.
        """

        return linkToResource(self.url, request, text, query_string)

    # Formatting-related functions.

    def formatText(self, text, request, fmt):

        """
        Format the given 'text' using the specified 'request' and formatter
        'fmt'.
        """

        # Assume plain text which is then formatted appropriately.

        return fmt.text(text)

class EventCalendar(EventResource):

    "An iCalendar resource."

    def __init__(self, url, calendar):
        EventResource.__init__(self, url)
        self.calendar = calendar
        self.events = None

    def getEvents(self):

        "Return a list of events from this resource."

        if self.events is None:
            self.events = []

            _calendar, _empty, calendar = self.calendar

            for objtype, attrs, obj in calendar:

                # Read events.

                if objtype == "VEVENT":
                    details = {}

                    for property, attrs, value in obj:

                        # Convert dates.

                        if property in ("DTSTART", "DTEND", "CREATED", "DTSTAMP", "LAST-MODIFIED"):
                            if property in ("DTSTART", "DTEND"):
                                property = property[2:]
                            if attrs.get("VALUE") == "DATE":
                                value = getDateFromCalendar(value)
                            else:
                                value = getDateTimeFromCalendar(value)

                        # Convert numeric data.

                        elif property == "SEQUENCE":
                            value = int(value)

                        # Convert lists.

                        elif property == "CATEGORIES":
                            value = to_list(value, ",")

                        # Convert positions (using decimal values).

                        elif property == "GEO":
                            try:
                                value = map(getMapReferenceFromDecimal, to_list(value, ";"))
                                if len(value) != 2:
                                    continue
                            except (KeyError, ValueError):
                                continue

                        # Accept other textual data as it is.

                        elif property in ("LOCATION", "SUMMARY", "URL"):
                            value = value or None

                        # Ignore other properties.

                        else:
                            continue

                        property = property.lower()
                        details[property] = value

                    self.events.append(CalendarEvent(self, details))

        return self.events

class EventPage:

    "An event page acting as an event resource."

    def __init__(self, page):
        self.page = page
        self.events = None
        self.body = None
        self.categories = None
        self.metadata = None

    def copyPage(self, page):

        "Copy the body of the given 'page'."

        self.body = page.getBody()

    def getPageURL(self):

        "Return the URL of this page."

        return getPageURL(self.page)

    def getFormat(self):

        "Get the format used on this page."

        return getFormat(self.page)

    def getMetadata(self):

        """
        Return a dictionary containing items describing the page's "created"
        time, "last-modified" time, "sequence" (or revision number) and the
        "last-comment" made about the last edit.
        """

        if self.metadata is None:
            self.metadata = getMetadata(self.page)
        return self.metadata

    def getRevisions(self):

        "Return a list of page revisions."

        return self.page.getRevList()

    def getPageRevision(self):

        "Return the revision details dictionary for this page."

        return getPageRevision(self.page)

    def getPageName(self):

        "Return the page name."

        return self.page.page_name

    def getPrettyPageName(self):

        "Return a nicely formatted title/name for this page."

        return getPrettyPageName(self.page)

    def getBody(self):

        "Get the current page body."

        if self.body is None:
            self.body = self.page.get_raw_body()
        return self.body

    def getEvents(self):

        "Return a list of events from this page."

        if self.events is None:
            details = {}
            self.events = [Event(self, details)]

            if self.getFormat() == "wiki":
                for match in definition_list_regexp.finditer(self.getBody()):

                    # Skip commented-out items.

                    if match.group("optcomment"):
                        continue

                    # Permit case-insensitive list terms.

                    term = match.group("term").lower()
                    desc = match.group("desc")

                    # Special value type handling.

                    # Dates.

                    if term in ("start", "end"):
                        desc = getDateTime(desc)

                    # Lists (whose elements may be quoted).

                    elif term in ("topics", "categories"):
                        desc = map(getSimpleWikiText, to_list(desc, ","))

                    # Position details.

                    elif term == "geo":
                        try:
                            desc = map(getMapReference, to_list(desc, None))
                            if len(desc) != 2:
                                continue
                        except (KeyError, ValueError):
                            continue

                    # Labels which may well be quoted.

                    elif term in ("title", "summary", "description", "location"):
                        desc = getSimpleWikiText(desc.strip())

                    if desc is not None:

                        # Handle apparent duplicates by creating a new set of
                        # details.

                        if details.has_key(term):

                            # Make a new event.

                            details = {}
                            self.events.append(Event(self, details))

                        details[term] = desc

        return self.events

    def setEvents(self, events):

        "Set the given 'events' on this page."

        self.events = events

    def getCategoryMembership(self):

        "Get the category names from this page."

        if self.categories is None:
            body = self.getBody()
            match = category_membership_regexp.search(body)
            self.categories = match and [x for x in match.groups() if x] or []

        return self.categories

    def setCategoryMembership(self, category_names):

        """
        Set the category membership for the page using the specified
        'category_names'.
        """

        self.categories = category_names

    def flushEventDetails(self):

        "Flush the current event details to this page's body text."

        new_body_parts = []
        end_of_last_match = 0
        body = self.getBody()

        events = iter(self.getEvents())

        event = events.next()
        event_details = event.getDetails()
        replaced_terms = set()

        for match in definition_list_regexp.finditer(body):

            # Permit case-insensitive list terms.

            term = match.group("term").lower()
            desc = match.group("desc")

            # Check that the term has not already been substituted. If so,
            # get the next event.

            if term in replaced_terms:
                try:
                    event = events.next()

                # No more events.

                except StopIteration:
                    break

                event_details = event.getDetails()
                replaced_terms = set()

            # Add preceding text to the new body.

            new_body_parts.append(body[end_of_last_match:match.start()])

            # Get the matching regions, adding the term to the new body.

            new_body_parts.append(match.group("wholeterm"))

            # Special value type handling.

            if event_details.has_key(term):

                # Dates.

                if term in ("start", "end"):
                    desc = desc.replace("YYYY-MM-DD", str(event_details[term]))

                # Lists (whose elements may be quoted).

                elif term in ("topics", "categories"):
                    desc = ", ".join([getEncodedWikiText(item) for item in event_details[term]])

                # Labels which must be quoted.

                elif term in ("title", "summary"):
                    desc = getEncodedWikiText(event_details[term])

                # Position details.

                elif term == "geo":
                    desc = " ".join(map(str, event_details[term]))

                # Text which need not be quoted, but it will be Wiki text.

                elif term in ("description", "link", "location"):
                    desc = event_details[term]

                replaced_terms.add(term)

            # Add the replaced value.

            new_body_parts.append(desc)

            # Remember where in the page has been processed.

            end_of_last_match = match.end()

        # Write the rest of the page.

        new_body_parts.append(body[end_of_last_match:])

        self.body = "".join(new_body_parts)

    def flushCategoryMembership(self):

        "Flush the category membership to the page body."

        body = self.getBody()
        category_names = self.getCategoryMembership()
        match = category_membership_regexp.search(body)

        if match:
            self.body = "".join([body[:match.start()], " ".join(category_names), body[match.end():]])

    def saveChanges(self):

        "Save changes to the event."

        self.flushEventDetails()
        self.flushCategoryMembership()
        self.page.saveText(self.getBody(), 0)

    def linkToPage(self, request, text, query_string=None):

        """
        Using 'request', return a link to this page with the given link 'text'
        and optional 'query_string'.
        """

        return linkToPage(request, self.page, text, query_string)

    # Formatting-related functions.

    def getParserClass(self, request, format):

        """
        Return a parser class using the 'request' for the given 'format', returning
        a plain text parser if no parser can be found for the specified 'format'.
        """

        try:
            return wikiutil.searchAndImportPlugin(request.cfg, "parser", format or "plain")
        except wikiutil.PluginMissingError:
            return wikiutil.searchAndImportPlugin(request.cfg, "parser", "plain")

    def formatText(self, text, request, fmt):

        """
        Format the given 'text' using the specified 'request' and formatter
        'fmt'.
        """

        fmt.page = self.page

        # Suppress line anchors.

        parser_cls = self.getParserClass(request, self.getFormat())
        parser = parser_cls(text, request, line_anchors=False)

        # Fix lists by indicating that a paragraph is already started.

        return request.redirectedOutput(parser.format, fmt, inhibit_p=True)

# Event details.

class Event(ActsAsTimespan):

    "A description of an event."

    def __init__(self, page, details):
        self.page = page
        self.details = details

        # Permit omission of the end of the event by duplicating the start.

        if self.details.has_key("start") and not self.details.get("end"):
            end = self.details["start"]

            # Make any end time refer to the day instead.

            if isinstance(end, DateTime):
                end = end.as_date()

            self.details["end"] = end

    def __repr__(self):
        return "<Event %r %r>" % (self.getSummary(), self.as_limits())

    def __hash__(self):

        """
        Return a dictionary hash, avoiding mistaken equality of events in some
        situations (notably membership tests) by including the URL as well as
        the summary.
        """

        return hash(self.getSummary() + self.getEventURL())

    def getPage(self):

        "Return the page describing this event."

        return self.page

    def setPage(self, page):

        "Set the 'page' describing this event."

        self.page = page

    def getEventURL(self):

        "Return the URL of this event."

        return self.page.getPageURL()

    def linkToEvent(self, request, text, query_string=None):

        """
        Using 'request', return a link to this event with the given link 'text'
        and optional 'query_string'.
        """

        return self.page.linkToPage(request, text, query_string)

    def getMetadata(self):

        """
        Return a dictionary containing items describing the event's "created"
        time, "last-modified" time, "sequence" (or revision number) and the
        "last-comment" made about the last edit.
        """

        # Delegate this to the page.

        return self.page.getMetadata()

    def getSummary(self, event_parent=None):

        """
        Return either the given title or summary of the event according to the
        event details, or a summary made from using the pretty version of the
        page name.

        If the optional 'event_parent' is specified, any page beneath the given
        'event_parent' page in the page hierarchy will omit this parent information
        if its name is used as the summary.
        """

        event_details = self.details

        if event_details.has_key("title"):
            return event_details["title"]
        elif event_details.has_key("summary"):
            return event_details["summary"]
        else:
            # If appropriate, remove the parent details and "/" character.

            title = self.page.getPageName()

            if event_parent and title.startswith(event_parent):
                title = title[len(event_parent.rstrip("/")) + 1:]

            return getPrettyTitle(title)

    def getDetails(self):

        "Return the details for this event."

        return self.details

    def setDetails(self, event_details):

        "Set the 'event_details' for this event."

        self.details = event_details

    # Timespan-related methods.

    def __contains__(self, other):
        return self == other

    def __eq__(self, other):
        if isinstance(other, Event):
            return self.getSummary() == other.getSummary() and self.getEventURL() == other.getEventURL() and self._cmp(other)
        else:
            return self._cmp(other) == 0

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self._cmp(other) == -1

    def __le__(self, other):
        return self._cmp(other) in (-1, 0)

    def __gt__(self, other):
        return self._cmp(other) == 1

    def __ge__(self, other):
        return self._cmp(other) in (0, 1)

    def _cmp(self, other):

        "Compare this event to an 'other' event purely by their timespans."

        if isinstance(other, Event):
            return cmp(self.as_timespan(), other.as_timespan())
        else:
            return cmp(self.as_timespan(), other)

    def as_timespan(self):
        details = self.details
        if details.has_key("start") and details.has_key("end"):
            return Timespan(details["start"], details["end"])
        else:
            return None

    def as_limits(self):
        ts = self.as_timespan()
        return ts and ts.as_limits()

class CalendarEvent(Event):

    "An event from a remote calendar."

    def getEventURL(self):

        "Return the URL of this event."

        return self.details.get("url") or self.page.getPageURL()

    def linkToEvent(self, request, text, query_string=None):

        """
        Using 'request', return a link to this event with the given link 'text'
        and optional 'query_string'.
        """

        return linkToResource(self.getEventURL(), request, text, query_string)

    def getMetadata(self):

        """
        Return a dictionary containing items describing the event's "created"
        time, "last-modified" time, "sequence" (or revision number) and the
        "last-comment" made about the last edit.
        """

        return {
            "created" : self.details.get("created") or self.details["dtstamp"],
            "last-modified" : self.details.get("last-modified") or self.details["dtstamp"],
            "sequence" : self.details.get("sequence") or 0,
            "last-comment" : ""
            }

# Obtaining event containers and events from such containers.

def getEventPages(pages):

    "Return a list of events found on the given 'pages'."

    # Get real pages instead of result pages.

    return map(EventPage, pages)

def getAllEventSources(request):

    "Return all event sources defined in the Wiki using the 'request'."

    sources_page = getattr(request.cfg, "event_aggregator_sources_page", "EventSourcesDict")

    # Remote sources are accessed via dictionary page definitions.

    return getWikiDict(sources_page, request)

def getEventResources(sources, calendar_start, calendar_end, request):

    """
    Return resource objects for the given 'sources' using the given
    'calendar_start' and 'calendar_end' to parameterise requests to the sources,
    and the 'request' to access configuration settings in the Wiki.
    """

    sources_dict = getAllEventSources(request)
    if not sources_dict:
        return []

    # Use dates for the calendar limits.

    if isinstance(calendar_start, Date):
        pass
    elif isinstance(calendar_start, Month):
        calendar_start = calendar_start.as_date(1)

    if isinstance(calendar_end, Date):
        pass
    elif isinstance(calendar_end, Month):
        calendar_end = calendar_end.as_date(-1)

    resources = []

    for source in sources:
        try:
            details = sources_dict[source].split()
            url = details[0]
            format = (details[1:] or ["ical"])[0]
        except (KeyError, ValueError):
            pass
        else:
            # Prevent local file access.

            if url.startswith("file:"):
                continue

            # Parameterise the URL.
            # Where other parameters are used, care must be taken to encode them
            # properly.

            url = url.replace("{start}", urllib.quote_plus(calendar_start and str(calendar_start) or ""))
            url = url.replace("{end}", urllib.quote_plus(calendar_end and str(calendar_end) or ""))

            # Get a parser.
            # NOTE: This could be done reactively by choosing a parser based on
            # NOTE: the content type provided by the URL.

            if format == "ical" and vCalendar is not None:
                parser = vCalendar.parse
                resource_cls = EventCalendar
                required_content_type = "text/calendar"
            else:
                continue

            # See if the URL is cached.

            cache_key = cache.key(request, content=url)
            cache_entry = caching.CacheEntry(request, "EventAggregator", cache_key, scope='wiki')

            # If no entry exists, or if the entry is older than a certain age
            # (5 minutes by default), create one with the response from the URL.

            now = time.time()
            mtime = cache_entry.mtime()
            max_cache_age = int(getattr(request.cfg, "event_aggregator_max_cache_age", "300"))

            # NOTE: The URL could be checked and the 'If-Modified-Since' header
            # NOTE: (see MoinMoin.action.pollsistersites) could be checked.

            if not cache_entry.exists() or now - mtime >= max_cache_age:

                # Access the remote data source.

                cache_entry.open(mode="w")

                try:
                    f = urllib2.urlopen(url)
                    try:
                        cache_entry.write(url + "\n")
                        cache_entry.write((f.headers.get("content-type") or "") + "\n")
                        cache_entry.write(f.read())
                    finally:
                        cache_entry.close()
                        f.close()

                # In case of an exception, just ignore the remote source.
                # NOTE: This could be reported somewhere.

                except IOError:
                    if cache_entry.exists():
                        cache_entry.remove()
                    continue

            # Open the cache entry and read it.

            cache_entry.open()
            try:
                data = cache_entry.read()
            finally:
                cache_entry.close()

            # Process the entry, parsing the content.

            f = StringIO(data)
            try:
                url = f.readline()

                # Get the content type and encoding, making sure that the data
                # can be parsed.

                content_type, encoding = getContentTypeAndEncoding(f.readline())
                if content_type != required_content_type:
                    continue

                # Send the data to the parser.

                uf = codecs.getreader(encoding or "utf-8")(f)
                try:
                    resources.append(resource_cls(url, parser(uf)))
                finally:
                    uf.close()
            finally:
                f.close()

    return resources

def getEventsFromResources(resources):

    "Return a list of events supplied by the given event 'resources'."

    events = []

    for resource in resources:

        # Get all events described by the resource.

        for event in resource.getEvents():

            # Remember the event.

            events.append(event)

    return events

# Event filtering and limits.

def getEventsInPeriod(events, calendar_period):

    """
    Return a collection containing those of the given 'events' which occur
    within the given 'calendar_period'.
    """

    all_shown_events = []

    for event in events:

        # Test for the suitability of the event.

        if event.as_timespan() is not None:

            # Compare the dates to the requested calendar window, if any.

            if event in calendar_period:
                all_shown_events.append(event)

    return all_shown_events

def getEventLimits(events):

    "Return the earliest and latest of the given 'events'."

    earliest = None
    latest = None

    for event in events:

        # Test for the suitability of the event.

        if event.as_timespan() is not None:
            ts = event.as_timespan()
            if earliest is None or ts.start < earliest:
                earliest = ts.start
            if latest is None or ts.end > latest:
                latest = ts.end

    return earliest, latest

def setEventTimestamps(request, events):

    """
    Using 'request', set timestamp details in the details dictionary of each of
    the 'events'.

    Return the latest timestamp found.
    """

    latest = None

    for event in events:
        event_details = event.getDetails()

        # Populate the details with event metadata.

        event_details.update(event.getMetadata())

        if latest is None or latest < event_details["last-modified"]:
            latest = event_details["last-modified"]

    return latest

def getOrderedEvents(events):

    """
    Return a list with the given 'events' ordered according to their start and
    end dates.
    """

    ordered_events = events[:]
    ordered_events.sort()
    return ordered_events

def getCalendarPeriod(calendar_start, calendar_end):

    """
    Return a calendar period for the given 'calendar_start' and 'calendar_end'.
    These parameters can be given as None.
    """

    # Re-order the window, if appropriate.

    if calendar_start is not None and calendar_end is not None and calendar_start > calendar_end:
        calendar_start, calendar_end = calendar_end, calendar_start

    return Timespan(calendar_start, calendar_end)

def getConcretePeriod(calendar_start, calendar_end, earliest, latest, resolution):

    """
    From the requested 'calendar_start' and 'calendar_end', which may be None,
    indicating that no restriction is imposed on the period for each of the
    boundaries, use the 'earliest' and 'latest' event months to define a
    specific period of interest.
    """

    # Define the period as starting with any specified start month or the
    # earliest event known, ending with any specified end month or the latest
    # event known.

    first = calendar_start or earliest
    last = calendar_end or latest

    # If there is no range of months to show, perhaps because there are no
    # events in the requested period, and there was no start or end month
    # specified, show only the month indicated by the start or end of the
    # requested period. If all events were to be shown but none were found show
    # the current month.

    if resolution == "date":
        get_current = getCurrentDate
    else:
        get_current = getCurrentMonth

    if first is None:
        first = last or get_current()
    if last is None:
        last = first or get_current()

    if resolution == "month":
        first = first.as_month()
        last = last.as_month()

    # Permit "expiring" periods (where the start date approaches the end date).

    return min(first, last), last

def getCoverage(events, resolution="date"):

    """
    Determine the coverage of the given 'events', returning a collection of
    timespans, along with a dictionary mapping locations to collections of
    slots, where each slot contains a tuple of the form (timespans, events).
    """

    all_events = {}
    full_coverage = TimespanCollection(resolution)

    # Get event details.

    for event in events:
        event_details = event.getDetails()

        # Find the coverage of this period for the event.

        # For day views, each location has its own slot, but for month
        # views, all locations are pooled together since having separate
        # slots for each location can lead to poor usage of vertical space.

        if resolution == "datetime":
            event_location = event_details.get("location")
        else:
            event_location = None

        # Update the overall coverage.

        full_coverage.insert_in_order(event)

        # Add a new events list for a new location.
        # Locations can be unspecified, thus None refers to all unlocalised
        # events.

        if not all_events.has_key(event_location):
            all_events[event_location] = [TimespanCollection(resolution, [event])]

        # Try and fit the event into an events list.

        else:
            slot = all_events[event_location]

            for slot_events in slot:

                # Where the event does not overlap with the events in the
                # current collection, add it alongside these events.

                if not event in slot_events:
                    slot_events.insert_in_order(event)
                    break

            # Make a new element in the list if the event cannot be
            # marked alongside existing events.

            else:
                slot.append(TimespanCollection(resolution, [event]))

    return full_coverage, all_events

def getCoverageScale(coverage):

    """
    Return a scale for the given coverage so that the times involved are
    exposed. The scale consists of a list of non-overlapping timespans forming
    a contiguous period of time.
    """

    times = set()
    for timespan in coverage:
        start, end = timespan.as_limits()

        # Add either genuine times or dates converted to times.

        if isinstance(start, DateTime):
            times.add(start)
        else:
            times.add(start.as_start_of_day())

        if isinstance(end, DateTime):
            times.add(end)
        else:
            times.add(end.as_date().next_day())

    times = list(times)
    times.sort(cmp_dates_as_day_start)

    scale = []
    first = 1
    start = None
    for time in times:
        if not first:
            scale.append(Timespan(start, time))
        else:
            first = 0
        start = time

    return scale

def getCountry(s):

    "Find a country code in the given string 's'."

    match = country_code_regexp.search(s)

    if match:
        return match.group("code")
    else:
        return None

# Page-related functions.

def fillEventPageFromTemplate(template_page, new_page, event_details, category_pagenames):

    """
    Using the given 'template_page', complete the 'new_page' by copying the
    template and adding the given 'event_details' (a dictionary of event
    fields), setting also the 'category_pagenames' to define category
    membership.
    """

    event_page = EventPage(template_page)
    new_event_page = EventPage(new_page)
    new_event_page.copyPage(event_page)

    if new_event_page.getFormat() == "wiki":
        new_event = Event(new_event_page, event_details)
        new_event_page.setEvents([new_event])
        new_event_page.setCategoryMembership(category_pagenames)
        new_event_page.flushEventDetails()

    return new_event_page.getBody()

# vim: tabstop=4 expandtab shiftwidth=4
